package com.compumark.gradle.plugin.ivy;

import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Stack;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.HeadMethod;
import org.apache.ivy.Ivy;
import org.apache.ivy.core.IvyPatternHelper;
import org.apache.ivy.core.deliver.DeliverOptions;
import org.apache.ivy.core.module.descriptor.Artifact;
import org.apache.ivy.core.module.descriptor.Configuration;
import org.apache.ivy.core.module.descriptor.DependencyDescriptor;
import org.apache.ivy.core.module.descriptor.ModuleDescriptor;
import org.apache.ivy.core.module.id.ModuleRevisionId;
import org.apache.ivy.core.publish.PublishOptions;
import org.apache.ivy.core.resolve.ResolvedModuleRevision;
import org.apache.ivy.plugins.resolver.DependencyResolver;
import org.apache.ivy.plugins.resolver.URLResolver;
import org.apache.jackrabbit.webdav.client.methods.DavMethod;
import org.apache.jackrabbit.webdav.client.methods.MkColMethod;
import org.gradle.api.GradleException;
import org.gradle.api.internal.ConventionTask;
import org.gradle.api.tasks.TaskAction;

public class Publish extends ConventionTask {
	private static final String PUBLISH_RESOLVER = "publish";
	
	public enum Status {
		INTEGRATION {
		    public String toString() { return "integration"; }			
		},
		RELEASE {
		    public String toString() { return "release"; }			
		}
	}
	
	private Ivy ivy = Ivy.newInstance();		
	private List<ModuleDescriptor> moduleDescriptors = new ArrayList<ModuleDescriptor>();
	private List<String> artifactsPatterns = new ArrayList<String>();
	private Status status = Status.INTEGRATION;
	private boolean validate = true;
	
	private boolean isTrueValue(String value) {
		 if (value == null) return false;
		 if (value.equalsIgnoreCase("on") || 
		     value.equalsIgnoreCase("yes") ||
		     value.equalsIgnoreCase("true") || 
		     value.equalsIgnoreCase("1")) {
			 return true;
		 }
		 return false;
	}

	@SuppressWarnings("unchecked")
	@TaskAction
	protected void publish() {
		artifactsPatterns.add(new File(getProject().getBuildDir(), "libs").getAbsolutePath() + "/[artifact]-[revision].[ext]");
		artifactsPatterns.add(new File(getProject().getBuildDir(), "distributions").getAbsolutePath() + "/[artifact]-[revision].[ext]");
		
		if (getProject().hasProperty("local")) {
    		System.out.println("Skipping task [local property set]");
    		return;
    	}
    	if (validate) validate();
    	
    	for (ModuleDescriptor moduleDescriptor : moduleDescriptors) {
	    	String revision = (String)getProject().property("version");
			String organisation = moduleDescriptor.getModuleRevisionId().getOrganisation();
			String module = moduleDescriptor.getModuleRevisionId().getName();
			String branch = moduleDescriptor.getModuleRevisionId().getBranch();
	    	
			ModuleRevisionId mrid = ModuleRevisionId.newInstance(organisation, module, branch, revision);
			ResolvedModuleRevision rmr = ivy.findModule(mrid);		
			if (rmr != null && rmr.getDescriptor().getStatus().equalsIgnoreCase(Status.RELEASE.toString())) {
				if (isTrueValue(System.getenv("IVY_FORCE_RELEASE"))) {
					System.out.println("Revision " + rmr +" has already been released, but force flag is set.");
				}
				else {
					throw new RuntimeException("Revision " + rmr +" has already been released.");
				}
			}
			
			String resolvedIvyPattern = getProject().getBuildDir().getAbsolutePath() + "/resolved-[module]-[revision].xml";
	    	try {
				DependencyResolver publishResolver = ivy.getSettings().getResolver(PUBLISH_RESOLVER);
				//ivy.getResolveEngine().setDictatorResolver(publishResolver);
	    		ivy.deliver(moduleDescriptor.getModuleRevisionId(), revision, resolvedIvyPattern, 
	    				new DeliverOptions()
	    					.setPubdate(new Date())
							.setStatus(status.toString())
							.setPubBranch(branch));
	    		
	    		if (publishResolver instanceof URLResolver){
	    			
	    			for(String ivyPattern : (List<String>)((URLResolver)publishResolver).getIvyPatterns()){
	    				mkdirs(IvyPatternHelper.substitute(ivyPattern, mrid));
	    			}
	    			Artifact[] artifacts = moduleDescriptor.getAllArtifacts();
	    			for (int i = 0; i < artifacts.length; i++) {
						Artifact artifact = artifacts[i];
						for(String artifactPattern : (List<String>)((URLResolver)publishResolver).getArtifactPatterns()){
							mkdirs(IvyPatternHelper.substitute(artifactPattern, mrid, artifact));
		    			}
					}
	    			
	    		}
	    		
	    		ivy.publish(moduleDescriptor.getModuleRevisionId(), artifactsPatterns, publishResolver.getName(), 
		        		new PublishOptions()
			        		.setSrcIvyPattern(resolvedIvyPattern)
			        		.setStatus(status.toString())
			        		.setValidate(true)
							.setPubrevision(revision)
							.setHaltOnMissing(true)
							.setUpdate(true)
							.setOverwrite(true)
							.setPubbranch(branch));
	    		
	    	}
	    	catch (Exception e) {
	    		throw new GradleException(e.getMessage(), e);
	    	}
    	}
    }
	
	/** Create path and all its parent folders, using MKCOL */
    private void mkdirs(String path) throws IOException {
        // Call mkdir on all parent path, starting at the topmost one
        final Stack<String> parents = new Stack<String>();
        path = getParentPath(path);
        while(!exists(path)) {
        	parents.push(path);
        	path = getParentPath(path);
        }
        
        while(!parents.isEmpty()) {
            mkdir(parents.pop());
        }
    }
    
    /** Return parent path: whatever comes before the last / in path, empty
     *  string if no / in path.
     */
    private String getParentPath(String path) {
        final int pos = path.lastIndexOf('/');
        if(pos > 0) {
            return path.substring(0, pos);
        } else {
            return "";
        }
    }
    
    /** Create path using MKCOL */
    private void mkdir(String path) throws IOException {
    	DavMethod mkcol = new MkColMethod(path);
    	new HttpClient().executeMethod(mkcol);
        if(mkcol.succeeded()) {
        	System.out.println(path+" CREATED");
        } else {
        	throw new IOException(mkcol.getStatusCode()+" "+mkcol.getStatusText());
        }
    }
    
    private boolean exists(String path) throws IOException {
        return new HttpClient().executeMethod(new HeadMethod(path)) == 200;
    }
   
	private void validate() {
		for (ModuleDescriptor moduleDescriptor : moduleDescriptors) {
			Configuration defaultConf = moduleDescriptor.getConfiguration("default");
			if (defaultConf == null) {
				throw new GradleException("Default configuration is missing!");
			}
	    	for (DependencyDescriptor dd : moduleDescriptor.getDependencies()) {
	    		ResolvedModuleRevision rmr = ivy.findModule(dd.getDependencyRevisionId());
	    		if (rmr == null) {
	    			throw new GradleException("Unable to find module " + dd.getDependencyRevisionId());
	    		}
	    		if (rmr.getDescriptor().getStatus().equalsIgnoreCase(Status.INTEGRATION.toString())) {
	    			throw new GradleException("No integration revisions are allowed. Dependency: "  + dd.getDependencyRevisionId());
	    		}
	    	}    	
			
		} 
		
    }
    
	public List<String> getArtifactsPatterns() {
		return artifactsPatterns;
	}
	
	public Ivy getIvy() {
		return ivy;
	}

	public void setIvy(Ivy ivy) {
		this.ivy = ivy;
	}

	public List<ModuleDescriptor> getModuleDescriptors() {
		return moduleDescriptors;
	}

	public void setModuleDescriptor(List<ModuleDescriptor> moduleDescriptors) {
		this.moduleDescriptors = moduleDescriptors;
	}

	public Status getStatus() {
		return status;
	}

	public void setStatus(Status status) {
		this.status = status;
	}

	public void setValidate(boolean validate) {
		this.validate = validate;
	}

	public boolean isValidate() {
		return validate;
	}
	
	public void addIvyFile(File ivyFile) throws ParseException, IOException {
		moduleDescriptors.add(ivy.resolve(ivyFile).getModuleDescriptor());
	}
}
